"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var shared_1 = require("./shared");
var TableErrs_1 = tslib_1.__importDefault(require("./TableErrs"));
var DecisionTable = (function () {
    function DecisionTable(id, factName, showLogs) {
        this.id = id;
        this._factName = factName;
        this.logger = new shared_1.Logger(showLogs);
        this.errs = new TableErrs_1.default(id);
        this.arrTable = [];
        this.imports = {};
        this.facts = [];
        this.conditions = [];
        this.actions = [];
    }
    Object.defineProperty(DecisionTable.prototype, "factName", {
        get: function () {
            return this._factName;
        },
        enumerable: false,
        configurable: true
    });
    DecisionTable.prototype.initTable = function (arrTable, facts, imports) {
        this.arrTable = arrTable;
        this.facts = facts;
        this.imports = imports;
        var colHeaderArr = shared_1.valsToArr(arrTable[0]);
        var opsStrArr = shared_1.valsToArr(arrTable[1]);
        var conditionsDone = false;
        this.conditions = [];
        this.actions = [];
        for (var i = 1; i < colHeaderArr.length; i++) {
            if (colHeaderArr[i] === 'Condition') {
                if (conditionsDone) {
                    throw Error(this.errs.colHeaderArgmt);
                }
                var condFunc = this.getCondOps(opsStrArr[i]);
                this.conditions.push(condFunc);
                conditionsDone = (colHeaderArr[i + 1] === 'Action');
            }
            else if (colHeaderArr[i] === 'Action') {
                if (!conditionsDone) {
                    throw Error(this.errs.colHeaderArgmt);
                }
                var actionFunc = this.getActionOps(opsStrArr[i]);
                this.actions.push(actionFunc);
                if (!colHeaderArr[i + 1]) {
                    break;
                }
            }
            else {
                throw Error(this.errs.colHeader);
            }
        }
    };
    DecisionTable.prototype.getCondOps = function (opStr) {
        var _this = this;
        var outer = this;
        return function (factIdx, paramVal) {
            var fact = outer.facts[factIdx];
            var errs = outer.errs;
            var arr = opStr.split(' ');
            var methodName = arr[0].replace('()', '');
            if (!opStr) {
                throw Error(errs.opBlank);
            }
            else if (arr.length !== 3) {
                throw Error(errs.opFormat + (" \"" + opStr + "\""));
            }
            else if (fact[methodName] === undefined) {
                throw Error(errs.attrUndef + (" \"" + opStr + "\""));
            }
            else if (arr[2] !== '$param') {
                throw Error(errs.mustEndWithParam + (" \"" + opStr + "\""));
            }
            var attrVal = null;
            if (typeof fact[methodName] === 'function') {
                attrVal = fact[methodName]();
            }
            else {
                attrVal = fact[methodName];
            }
            return _this.compareVals(arr[1], attrVal, paramVal);
        };
    };
    DecisionTable.prototype.compareVals = function (operator, val1, val2) {
        if (operator === '===') {
            return val1 === val2;
        }
        else if (operator === '==') {
            return val1 === val2;
        }
        else if (operator === '!=') {
            return val1 !== val2;
        }
        else if (operator === '!==') {
            return val1 !== val2;
        }
        else if (operator === '>') {
            return val1 > val2;
        }
        else if (operator === '>=') {
            return val1 >= val2;
        }
        else if (operator === '<') {
            return val1 < val2;
        }
        else if (operator === '<=') {
            return val1 <= val2;
        }
        else {
            throw Error(this.errs.notAnOp + (" '" + operator + "'"));
        }
    };
    DecisionTable.prototype.getActionOps = function (actionStr) {
        if (!actionStr) {
            throw Error(this.errs.opBlank);
        }
        var outer = this;
        var errs = this.errs;
        return function (factIdx, cellVals) {
            var argLength = actionStr.split('$param').length - 1;
            var op = " \"" + actionStr + "\"";
            if (argLength !== cellVals.length) {
                throw Error(errs.paramCount + op);
            }
            var opArr = actionStr.split(' ');
            var fact = outer.facts[factIdx];
            if (opArr[1] === '=') {
                if (cellVals.length !== 1) {
                    throw Error(errs.assignParamCount + op);
                }
                else if (fact[opArr[0]] === undefined) {
                    throw Error(errs.attrUndef + op);
                }
                fact[opArr[0]] = cellVals[0];
            }
            else {
                var n = actionStr.lastIndexOf('(');
                var methodName = actionStr.substring(0, n);
                if (fact[methodName] === undefined) {
                    throw Error(errs.attrUndef + op);
                }
                fact[methodName].apply(fact, cellVals);
            }
        };
    };
    DecisionTable.prototype.updateFacts = function () {
        for (var factIdx = 0; factIdx < this.facts.length; factIdx++) {
            rowLoop: for (var rowIdx = 2; rowIdx < this.arrTable.length; rowIdx++) {
                var ruleArr = shared_1.valsToArr(this.arrTable[rowIdx]);
                if (ruleArr[0] === '') {
                    throw Error(this.errs.ruleNameEmpty);
                }
                var colIdx = 1;
                for (var i = 0; i < this.conditions.length; i++) {
                    var passed = this.callCondOp(factIdx, i, ruleArr[colIdx++]);
                    if (!passed) {
                        continue rowLoop;
                    }
                }
                for (var i = 0; i < this.actions.length; i++) {
                    this.callActionOp(factIdx, i, ruleArr[colIdx++]);
                }
            }
        }
        return this.facts;
    };
    DecisionTable.prototype.callCondOp = function (factIdx, condIdx, cellValStr) {
        if (cellValStr === '') {
            return true;
        }
        var retVal = shared_1.parseCell(cellValStr, this.imports);
        if (retVal === null) {
            throw Error(this.errs.invalidVal + (" '" + cellValStr + "'"));
        }
        return this.conditions[condIdx](factIdx, retVal);
    };
    DecisionTable.prototype.callActionOp = function (factIdx, actionIdx, cellValStr) {
        if (cellValStr === '') {
            return;
        }
        var cellVals = cellValStr.split(',');
        for (var i = 0; i < cellVals.length; i++) {
            var val = shared_1.parseCell(cellVals[i], this.imports);
            if (val === null) {
                throw Error(this.errs.invalidVal + (" '" + cellValStr + "'"));
            }
            else {
                cellVals[i] = val;
            }
        }
        this.actions[actionIdx](factIdx, cellVals);
    };
    return DecisionTable;
}());
exports.default = DecisionTable;
//# sourceMappingURL=DecisionTable.js.map