"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var csvtojson_1 = tslib_1.__importDefault(require("csvtojson"));
var shared_1 = require("./shared");
var DecisionTable_1 = tslib_1.__importDefault(require("./DecisionTable"));
var TableErrs_1 = tslib_1.__importDefault(require("./TableErrs"));
var Trool = (function () {
    function Trool() {
        this.NO_TABLES_WARN = 'No decision tables found';
        this.IMPORT_START_ERR = 'Import start format error for';
        this.IMPORT_PROP_ERR = 'Import property can only be alpha-numeric and underscores ';
        this.UPDATE_START_MSG = ' DecisionTables found. Applying table logic to facts.';
        this.IMPORT_NAME_WARN = '!!WARNING!! The spreadsheet is using an import name ' +
            'already passed via the imports object. The spreadsheet will overwrite the import: ';
        this.logger = new shared_1.Logger();
        this._decisionTables = [];
    }
    Trool.prototype.init = function (filePath, facts, showLogs, imports) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var jsonArr;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, csvtojson_1.default().fromFile(filePath)];
                    case 1:
                        jsonArr = _a.sent();
                        this.createDecisionTables(jsonArr, facts, showLogs, imports);
                        return [2];
                }
            });
        });
    };
    Trool.prototype.initFromString = function (fileContent, facts, showLogs, imports) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var jsonArr;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, csvtojson_1.default().fromString(fileContent)];
                    case 1:
                        jsonArr = _a.sent();
                        this.createDecisionTables(jsonArr, facts, showLogs, imports);
                        return [2];
                }
            });
        });
    };
    Trool.prototype.createDecisionTables = function (jsonArr, facts, showLogs, imports) {
        this.logger = new shared_1.Logger(showLogs);
        var allImports = this.setupImports(jsonArr, imports || {});
        this._decisionTables = this.getTables(jsonArr, facts, allImports);
    };
    Trool.prototype.applyRules = function () {
        var tableCount = this._decisionTables.length;
        if (tableCount === 0) {
            this.logger.warn(this.NO_TABLES_WARN);
            return {};
        }
        else {
            this.logger.log(tableCount + this.UPDATE_START_MSG);
        }
        var updatedFacts = {};
        for (var i = 0; i < tableCount; i++) {
            var table = this._decisionTables[i];
            updatedFacts[table.factName] = table.updateFacts();
        }
        return updatedFacts;
    };
    Object.defineProperty(Trool.prototype, "decisionTables", {
        get: function () {
            return this._decisionTables;
        },
        enumerable: false,
        configurable: true
    });
    Trool.prototype.setupImports = function (jsonArr, imports) {
        var importName = '';
        var newImportObj = {};
        for (var i = 0; i < jsonArr.length; i++) {
            var firstCell = jsonArr[i].field1.trim();
            if (firstCell.startsWith('Import:')) {
                importName = this.getImportName(firstCell, imports);
            }
            else if (!!importName) {
                if (!/^[a-zA-Z0-9-_]+$/.test(firstCell)) {
                    throw Error(this.IMPORT_PROP_ERR + firstCell);
                }
                newImportObj[firstCell] = shared_1.parseCell(jsonArr[i].field2, imports);
                if (this.isLastRow(jsonArr, i)) {
                    imports[importName] = newImportObj;
                    importName = '';
                    newImportObj = {};
                }
            }
        }
        return imports;
    };
    Trool.prototype.getImportName = function (firstCell, imports) {
        var firstCellArr = firstCell.split(' ');
        if (firstCellArr.length !== 2) {
            throw Error(this.IMPORT_START_ERR + (" '" + firstCell + "'"));
        }
        var importName = firstCellArr[1];
        if (imports.hasOwnProperty(importName)) {
            this.logger.warn(this.IMPORT_NAME_WARN + importName);
        }
        return importName;
    };
    Trool.prototype.getTables = function (jsonArr, facts, imports) {
        var decisionTables = [];
        var startCellArr = null;
        var tableStart = -1;
        for (var i = 0; i < jsonArr.length; i++) {
            var firstCol = jsonArr[i].field1.trim();
            if (firstCol.startsWith('Table:')) {
                tableStart = i;
                startCellArr = firstCol.split(' ');
            }
            else if (startCellArr && this.isLastRow(jsonArr, i)) {
                var showLogs = this.logger.showLogs;
                var id = decisionTables.length + 1;
                var tableRows = jsonArr.slice(tableStart, i + 1);
                var factArr = this.getFacts(startCellArr, id, facts);
                var table = new DecisionTable_1.default(id, startCellArr[1], showLogs);
                table.initTable(tableRows, factArr, imports);
                decisionTables.push(table);
                tableStart = -1;
                startCellArr = null;
            }
        }
        return decisionTables;
    };
    Trool.prototype.getFacts = function (startCellArr, id, facts) {
        var tableErrs = new TableErrs_1.default(id);
        if (startCellArr.length !== 2) {
            throw Error(tableErrs.startCell);
        }
        else if (!facts[startCellArr[1]]) {
            throw Error(tableErrs.factFalsey);
        }
        var factArr = facts[startCellArr[1]];
        return (factArr instanceof Array) ? factArr : [factArr];
    };
    Trool.prototype.isLastRow = function (jsonArr, idx) {
        var nextCell = jsonArr[idx + 1] ? jsonArr[idx + 1].field1.trim() : '';
        return !nextCell || nextCell.startsWith('Table:') || nextCell.startsWith('Import:');
    };
    return Trool;
}());
exports.default = Trool;
//# sourceMappingURL=Trool.js.map